#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(bmp180);

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/i2c.h>
#include <stdbool.h>

#include <math.h>

#include "bmp180.h"
#include "sds011.h"

#define BMP180_I2CADDR 0x77 //!< BMP180 I2C address

#define BMP180_ULTRALOWPOWER 0 //!< Ultra low power mode
#define BMP180_STANDARD 1      //!< Standard mode
#define BMP180_HIGHRES 2       //!< High-res mode
#define BMP180_ULTRAHIGHRES 3  //!< Ultra high-res mode

#define OVERSAMPLING BMP180_ULTRALOWPOWER

#define BMP180_CAL_AC1 0xAA    //!< R   Calibration data (16 bits)
#define BMP180_CAL_AC2 0xAC    //!< R   Calibration data (16 bits)
#define BMP180_CAL_AC3 0xAE    //!< R   Calibration data (16 bits)
#define BMP180_CAL_AC4 0xB0    //!< R   Calibration data (16 bits)
#define BMP180_CAL_AC5 0xB2    //!< R   Calibration data (16 bits)
#define BMP180_CAL_AC6 0xB4    //!< R   Calibration data (16 bits)
#define BMP180_CAL_B1 0xB6     //!< R   Calibration data (16 bits)
#define BMP180_CAL_B2 0xB8     //!< R   Calibration data (16 bits)
#define BMP180_CAL_MB 0xBA     //!< R   Calibration data (16 bits)
#define BMP180_CAL_MC 0xBC     //!< R   Calibration data (16 bits)
#define BMP180_CAL_MD 0xBE     //!< R   Calibration data (16 bits)

#define BMP180_CONTROL 0xF4         //!< Control register
#define BMP180_TEMPDATA 0xF6        //!< Temperature data register
#define BMP180_PRESSUREDATA 0xF6    //!< Pressure data register
#define BMP180_READTEMPCMD 0x2E     //!< Read temperature control register value
#define BMP180_READPRESSURECMD 0x34 //!< Read pressure control register value

#define I2C_PORT DT_NODELABEL(i2c0)

int16_t ac1, ac2, ac3, b1, b2, mb, mc, md;
uint16_t ac4, ac5, ac6;

const struct device *devi2c;

uint8_t read_reg_8(const struct device *devi2c, uint8_t reg)
{
  uint8_t bufin[1];
  uint8_t bufout[1];
  int ret;
  bufin[0]=reg;
  ret = i2c_write_read(devi2c, 0x77, bufin, 1, bufout, 1);
  if (ret < 0) {
      printk("Failed to read from devi2cice\n");
      return ret;
  }
  return bufout[0];
}

uint16_t read_reg_16(const struct device *devi2c, uint8_t reg)
{
  uint16_t bufout;
  int ret;
  ret = i2c_write_read(devi2c, BMP180_I2CADDR, &reg, 1, &bufout, 2);
//  ret = i2c_burst_read(devi2c, BMP180_I2CADDR, reg, 
  if (ret < 0) {
      printk("Failed to read from devi2cice\n");
      return ret;
  //} else {
  //  printk("Reg %d Value %d\n", reg, sys_be16_to_cpu(bufout));
}
  return sys_be16_to_cpu(bufout);
}

int write_reg_8(const struct device *devi2c, uint8_t reg, uint8_t data)
{
  int ret;
  ret = i2c_reg_write_byte(devi2c, BMP180_I2CADDR, reg, data);
  if (ret < 0) {
      printk("Failed to write to device\n");
      return ret;
  }
  return ret;
}

int32_t computeB5(int32_t UT) {
  int32_t X1 = (UT - (int32_t)ac6) * ((int32_t)ac5) >> 15;
  int32_t X2 = ((int32_t)mc << 11) / (X1 + (int32_t)md);
  return X1 + X2;
}

uint16_t readRawTemperature(const struct device *devi2c)
{
  write_reg_8(devi2c, BMP180_CONTROL, BMP180_READTEMPCMD);
  k_msleep(10);
  return read_reg_16(devi2c, BMP180_TEMPDATA);
}

uint16_t readRawPressure(const struct device *devi2c)
{  
  uint32_t raw;
  /*uint8_t raw1;
  uint8_t raw2;
  uint8_t raw3;*/

  write_reg_8(devi2c, BMP180_CONTROL, BMP180_READPRESSURECMD + (OVERSAMPLING << 6));

  if (OVERSAMPLING == BMP180_ULTRALOWPOWER) {
    k_msleep(5);
  } else if (OVERSAMPLING == BMP180_STANDARD) {
    k_msleep(8);
  } else if (OVERSAMPLING == BMP180_HIGHRES) {
    k_msleep(14);
  } else {
    k_msleep(26);
  }
   raw = read_reg_16(devi2c, BMP180_PRESSUREDATA);

  raw <<= 8;
  raw |= read_reg_8(devi2c, BMP180_PRESSUREDATA + 2);
  
  
/*  raw1 = read_reg_8(devi2c, BMP180_PRESSUREDATA);
  raw2 = read_reg_8(devi2c, BMP180_PRESSUREDATA + 1);
  raw3 = read_reg_8(devi2c, BMP180_PRESSUREDATA + 2);
  
  	raw = raw1 << 16 | raw2 << 8 | raw3;

  */
  raw >>= (8 - OVERSAMPLING);

  return raw;
}

int32_t readPressure(const struct device *devi2c) {
  int32_t UT, UP, B3, B5, B6, X1, X2, X3, p;
  uint32_t B4, B7;

  UT = readRawTemperature(devi2c);
  UP = readRawPressure(devi2c);

  B5 = computeB5(UT);
/* 
  // do pressure calcs
  B6 = B5 - 4000;
  X1 = ((int32_t)b2 * ((B6 * B6) >> 12)) >> 11;
  X2 = ((int32_t)ac2 * B6) >> 11;
  X3 = X1 + X2;
  B3 = ((((int32_t)ac1 * 4 + X3) << OVERSAMPLING) + 2) / 4;

  X1 = ((int32_t)ac3 * B6) >> 13;
  X2 = ((int32_t)b1 * ((B6 * B6) >> 12)) >> 16;
  X3 = ((X1 + X2) + 2) >> 2;
  B4 = ((uint32_t)ac4 * (uint32_t)(X3 + 32768)) >> 15;
  B7 = ((uint32_t)UP - B3) * (uint32_t)(50000UL >> OVERSAMPLING);

  if (B7 < 0x80000000) {
    p = (B7 * 2) / B4;
  } else {
    p = (B7 / B4) * 2;
  }
  X1 = (p >> 8) * (p >> 8);
  X1 = (X1 * 3038) >> 16;
  X2 = (-7357 * p) >> 16;

  p = p + ((X1 + X2 + (int32_t)3791) >> 4);
  */
  B6= B5 - 4000;
	X1 = (b2 * ((B6 * B6) >> 12)) >> 11;
	X2 = (ac2 * B6) >> 11;
	X3 = X1 + X2;
	B3 = (((ac1 * 4 + X3) << OVERSAMPLING) + 2) >> 2;

	X1 = (ac3 * B6) >> 13;
	X2 = (b1 * ((B6 * B6) >> 12)) >> 16;
	X3 = ((X1 + X2) + 2) >> 2;
	B4 = (ac4 * (unsigned long)(X3 + 32768UL)) >> 15;
	B7 = ((unsigned long)UP - B3) * (50000UL >> OVERSAMPLING);

	if(B7 < 0x80000000)
		p = (B7 * 2) / B4;
	else
		p = (B7 / B4) * 2;
	
	X1 = (p >> 8) * (p >> 8);
	X1 = (X1 * 3038) >> 16;
	X2 = (-7357 * p) >> 16;

	p = p + ((X1 + X2 + 3791) >> 4);
  
  return p;
}
 
uint32_t readSealevelPressure(const struct device *devi2c, float altitude_meters) {
  float pressure = readPressure(devi2c);
  return (uint32_t)(pressure / pow(1.0 - altitude_meters / 44330, 5.255));
};

uint32_t bmp180_readPressure() {
  return readSealevelPressure(devi2c, 146.0);
};
 
int bmp180_setup()
{
  int ret;
  devi2c = DEVICE_DT_GET(I2C_PORT);

  if (!devi2c) {
      printk("I2C: device driver not found.\n");
      return -1;
  }

  ret = write_reg_8(devi2c, 0xe0, 0xb6);
  if(ret < 0) {
    return ret;
  }
  k_msleep(500);
  
  ac1 = read_reg_16(devi2c, BMP180_CAL_AC1);
  ac2 = read_reg_16(devi2c, BMP180_CAL_AC2);
  ac3 = read_reg_16(devi2c, BMP180_CAL_AC3);
  ac4 = read_reg_16(devi2c, BMP180_CAL_AC4);
  ac5 = read_reg_16(devi2c, BMP180_CAL_AC5);
  ac6 = read_reg_16(devi2c, BMP180_CAL_AC6);

  b1 = read_reg_16(devi2c, BMP180_CAL_B1);
  b2 = read_reg_16(devi2c, BMP180_CAL_B2);

  mb = read_reg_16(devi2c, BMP180_CAL_MB);
  mc = read_reg_16(devi2c, BMP180_CAL_MC);
  md = read_reg_16(devi2c, BMP180_CAL_MD);

  return 0;
};

